@@ -20,13 +20,17 @@ from .http_stream import HTTPStream
 from .ws_stream import WSStream
 from ..config import Config
 from ..events import Closed, Event, RawData, Updated
-from ..typing import Event as IOEvent
+from ..typing import ASGIFramework, Context, Event as IOEvent
 from ..utils import filter_pseudo_headers
 
 BUFFER_HIGH_WATER = 2 * 2 ** 14  # Twice the default max frame size (two frames worth)
 BUFFER_LOW_WATER = BUFFER_HIGH_WATER / 2
 
 
+class BufferCompleteError(Exception):
+    pass
+
+
 class StreamBuffer:
     def __init__(self, event_class: Type[IOEvent]) -> None:
         self.buffer = bytearray()
@@ -52,7 +56,7 @@ class StreamBuffer:
 
     async def push(self, data: bytes) -> None:
         if self._complete:
-            raise RuntimeError("Push to completed StreamBuffer")
+            raise BufferCompleteError()
         self.buffer.extend(data)
         await self._is_empty.clear()
         if len(self.buffer) >= BUFFER_HIGH_WATER:
@@ -73,17 +77,19 @@ class StreamBuffer:
 class H2Protocol:
     def __init__(
         self,
+        app: ASGIFramework,
         config: Config,
+        context: Context,
         ssl: bool,
         client: Optional[Tuple[str, int]],
         server: Optional[Tuple[str, int]],
         send: Callable[[Event], Awaitable[None]],
-        spawn_app: Callable[[dict, Callable], Awaitable[Callable]],
-        event_class: Type[IOEvent],
     ) -> None:
+        self.app = app
         self.client = client
         self.closed = False
         self.config = config
+        self.context = context
 
         self.connection = h2.connection.H2Connection(
             config=h2.config.H2Configuration(client_side=False, header_encoding=None)
@@ -98,14 +104,12 @@ class H2Protocol:
             },
         )
 
-        self.event_class = event_class
         self.send = send
         self.server = server
-        self.spawn_app = spawn_app
         self.ssl = ssl
         self.streams: Dict[int, Union[HTTPStream, WSStream]] = {}
         # The below are used by the sending task
-        self.has_data = event_class()
+        self.has_data = self.context.event_class()
         self.priority = priority.PriorityTree()
         self.stream_buffers: Dict[int, StreamBuffer] = {}
 
@@ -181,6 +185,7 @@ class H2Protocol:
             stream_ids = list(self.streams.keys())
             for stream_id in stream_ids:
                 await self._close_stream(stream_id)
+            await self.has_data.set()
 
     async def stream_send(self, event: StreamEvent) -> None:
         try:
@@ -206,7 +211,12 @@ class H2Protocol:
                 await self.send(Updated())
             elif isinstance(event, Request):
                 await self._create_server_push(event.stream_id, event.raw_path, event.headers)
-        except h2.exceptions.ProtocolError:
+        except (
+            BufferCompleteError,
+            KeyError,
+            priority.MissingStreamError,
+            h2.exceptions.ProtocolError,
+        ):
             # Connection has closed whilst blocked on flow control or
             # connection has advanced ahead of the last emitted event.
             return
@@ -280,25 +290,27 @@ class H2Protocol:
 
         if method == "CONNECT":
             self.streams[request.stream_id] = WSStream(
+                self.app,
                 self.config,
+                self.context,
                 self.ssl,
                 self.client,
                 self.server,
                 self.stream_send,
-                self.spawn_app,
                 request.stream_id,
             )
         else:
             self.streams[request.stream_id] = HTTPStream(
+                self.app,
                 self.config,
+                self.context,
                 self.ssl,
                 self.client,
                 self.server,
                 self.stream_send,
-                self.spawn_app,
                 request.stream_id,
             )
-        self.stream_buffers[request.stream_id] = StreamBuffer(self.event_class)
+        self.stream_buffers[request.stream_id] = StreamBuffer(self.context.event_class)
         try:
             self.priority.insert_stream(request.stream_id)
         except priority.DuplicateStreamError:
@@ -346,3 +358,4 @@ class H2Protocol:
         if stream_id in self.streams:
             stream = self.streams.pop(stream_id)
             await stream.handle(StreamClosed(stream_id=stream_id))
+            await self.has_data.set()
